@@ -35,6 +35,7 @@
 #include "lib/ovn-util.h"
 #include "memory.h"
 #include "openvswitch/poll-loop.h"
+#include "ovsdb-idl.h"
 #include "simap.h"
 #include "smap.h"
 #include "sset.h"
@@ -654,6 +655,11 @@ sync_remote_port(struct ic_context *ctx,
     /* Sync tunnel key from ISB to NB */
     sync_lsp_tnl_key(lsp, isb_pb->tunnel_key);
 
+    /* Skip port binding if it is already requested by the CMS. */
+    if (smap_get(&lsp->options, "requested-chassis")) {
+        return;
+    }
+
     /* Sync gateway from ISB to SB */
     if (isb_pb->gateway[0]) {
         if (!sb_pb->chassis || strcmp(sb_pb->chassis->name, isb_pb->gateway)) {
@@ -869,6 +875,8 @@ struct ic_route_info {
     const char *origin;
     const char *route_table;
 
+    const struct nbrec_logical_router *nb_lr;
+
     /* Either nb_route or nb_lrp is set and the other one must be NULL.
      * - For a route that is learned from IC-SB, or a static route that is
      *   generated from a route that is configured in NB, the "nb_route"
@@ -945,7 +953,8 @@ parse_route(const char *s_prefix, const char *s_nexthop,
 /* Return false if can't be added due to bad format. */
 static bool
 add_to_routes_learned(struct hmap *routes_learned,
-                      const struct nbrec_logical_router_static_route *nb_route)
+                      const struct nbrec_logical_router_static_route *nb_route,
+                      const struct nbrec_logical_router *nb_lr)
 {
     struct in6_addr prefix, nexthop;
     unsigned int plen;
@@ -967,6 +976,7 @@ add_to_routes_learned(struct hmap *routes_learned,
     ic_route->nb_route = nb_route;
     ic_route->origin = origin;
     ic_route->route_table = nb_route->route_table;
+    ic_route->nb_lr = nb_lr;
     hmap_insert(routes_learned, &ic_route->node,
                 ic_route_hash(&prefix, plen, &nexthop, origin,
                               nb_route->route_table));
@@ -1110,7 +1120,8 @@ add_to_routes_ad(struct hmap *routes_ad, const struct in6_addr prefix,
                  unsigned int plen, const struct in6_addr nexthop,
                  const char *origin, const char *route_table,
                  const struct nbrec_logical_router_port *nb_lrp,
-                 const struct nbrec_logical_router_static_route *nb_route)
+                 const struct nbrec_logical_router_static_route *nb_route,
+                 const struct nbrec_logical_router *nb_lr)
 {
     if (route_table == NULL) {
         route_table = "";
@@ -1128,6 +1139,7 @@ add_to_routes_ad(struct hmap *routes_ad, const struct in6_addr prefix,
         ic_route->origin = origin;
         ic_route->route_table = route_table;
         ic_route->nb_lrp = nb_lrp;
+        ic_route->nb_lr = nb_lr;
         hmap_insert(routes_ad, &ic_route->node, hash);
     } else {
         static struct vlog_rate_limit rl = VLOG_RATE_LIMIT_INIT(5, 1);
@@ -1141,6 +1153,7 @@ static void
 add_static_to_routes_ad(
     struct hmap *routes_ad,
     const struct nbrec_logical_router_static_route *nb_route,
+    const struct nbrec_logical_router *nb_lr,
     const struct lport_addresses *nexthop_addresses,
     const struct smap *nb_options)
 {
@@ -1183,14 +1196,15 @@ add_static_to_routes_ad(
     }
 
     add_to_routes_ad(routes_ad, prefix, plen, nexthop, ROUTE_ORIGIN_STATIC,
-                     nb_route->route_table, NULL, nb_route);
+                     nb_route->route_table, NULL, nb_route, nb_lr);
 }
 
 static void
 add_network_to_routes_ad(struct hmap *routes_ad, const char *network,
                          const struct nbrec_logical_router_port *nb_lrp,
                          const struct lport_addresses *nexthop_addresses,
-                         const struct smap *nb_options)
+                         const struct smap *nb_options,
+                         const struct nbrec_logical_router *nb_lr)
 {
     struct in6_addr prefix, nexthop;
     unsigned int plen;
@@ -1229,7 +1243,7 @@ add_network_to_routes_ad(struct hmap *routes_ad, const char *network,
 
     /* directly-connected routes go to <main> route table */
     add_to_routes_ad(routes_ad, prefix, plen, nexthop, ROUTE_ORIGIN_CONNECTED,
-                     NULL, nb_lrp, NULL);
+                     NULL, nb_lrp, NULL, nb_lr);
 }
 
 static bool
@@ -1333,7 +1347,6 @@ lrp_is_ts_port(struct ic_context *ctx, struct ic_router_info *ic_lr,
 
 static void
 sync_learned_routes(struct ic_context *ctx,
-                    const struct icsbrec_availability_zone *az,
                     struct ic_router_info *ic_lr)
 {
     ovs_assert(ctx->ovnnb_txn);
@@ -1356,7 +1369,15 @@ sync_learned_routes(struct ic_context *ctx,
 
         ICSBREC_ROUTE_FOR_EACH_EQUAL (isb_route, isb_route_key,
                                       ctx->icsbrec_route_by_ts) {
-            if (isb_route->availability_zone == az) {
+            const char *lr_id = smap_get(&isb_route->external_ids, "lr-id");
+            if (lr_id == NULL) {
+                continue;
+            }
+            struct uuid lr_uuid;
+            if (!uuid_from_string(&lr_uuid, lr_id)) {
+                continue;
+            }
+            if (uuid_equals(&ic_lr->lr->header_.uuid, &lr_uuid)) {
                 continue;
             }
 
@@ -1446,16 +1467,24 @@ static void
 ad_route_sync_external_ids(const struct ic_route_info *route_adv,
                            const struct icsbrec_route *isb_route)
 {
-    struct uuid isb_ext_id, nb_id;
+    struct uuid isb_ext_id, nb_id, isb_ext_lr_id, lr_id;
     smap_get_uuid(&isb_route->external_ids, "nb-id", &isb_ext_id);
+    smap_get_uuid(&isb_route->external_ids, "lr-id", &isb_ext_lr_id);
     nb_id = route_adv->nb_route ? route_adv->nb_route->header_.uuid
                                : route_adv->nb_lrp->header_.uuid;
+    lr_id = route_adv->nb_lr->header_.uuid;
     if (!uuid_equals(&isb_ext_id, &nb_id)) {
         char *uuid_s = xasprintf(UUID_FMT, UUID_ARGS(&nb_id));
         icsbrec_route_update_external_ids_setkey(isb_route, "nb-id",
                                                  uuid_s);
         free(uuid_s);
     }
+    if (!uuid_equals(&isb_ext_lr_id, &lr_id)) {
+        char *uuid_s = xasprintf(UUID_FMT, UUID_ARGS(&lr_id));
+        icsbrec_route_update_external_ids_setkey(isb_route, "lr-id",
+                                                 uuid_s);
+        free(uuid_s);
+    }
 }
 
 /* Sync routes from routes_ad to IC-SB. */
@@ -1560,7 +1589,7 @@ build_ts_routes_to_adv(struct ic_context *ctx,
         if (smap_get_uuid(&nb_route->external_ids, "ic-learned-route",
                           &isb_uuid)) {
             /* It is a learned route */
-            if (!add_to_routes_learned(&ic_lr->routes_learned, nb_route)) {
+            if (!add_to_routes_learned(&ic_lr->routes_learned, nb_route, lr)) {
                 static struct vlog_rate_limit rl = VLOG_RATE_LIMIT_INIT(5, 1);
                 VLOG_WARN_RL(&rl, "Bad format of learned route in NB: "
                              "%s -> %s. Delete it.", nb_route->ip_prefix,
@@ -1570,7 +1599,7 @@ build_ts_routes_to_adv(struct ic_context *ctx,
             }
         } else if (!strcmp(ts_route_table, nb_route->route_table)) {
             /* It may be a route to be advertised */
-            add_static_to_routes_ad(routes_ad, nb_route, ts_port_addrs,
+            add_static_to_routes_ad(routes_ad, nb_route, lr, ts_port_addrs,
                                     &nb_global->options);
         }
     }
@@ -1582,7 +1611,8 @@ build_ts_routes_to_adv(struct ic_context *ctx,
             for (int j = 0; j < lrp->n_networks; j++) {
                 add_network_to_routes_ad(routes_ad, lrp->networks[j], lrp,
                                          ts_port_addrs,
-                                         &nb_global->options);
+                                         &nb_global->options,
+                                         lr);
             }
         } else {
             /* The router port of the TS port is ignored. */
@@ -1593,9 +1623,9 @@ build_ts_routes_to_adv(struct ic_context *ctx,
 }
 
 static void
-advertise_lr_routes(struct ic_context *ctx,
-                    const struct icsbrec_availability_zone *az,
-                    struct ic_router_info *ic_lr)
+collect_lr_routes(struct ic_context *ctx,
+                  struct ic_router_info *ic_lr,
+                  struct shash *routes_ad_by_ts)
 {
     const struct nbrec_nb_global *nb_global =
         nbrec_nb_global_first(ctx->ovnnb_idl);
@@ -1606,15 +1636,26 @@ advertise_lr_routes(struct ic_context *ctx,
     struct lport_addresses ts_port_addrs;
     const struct icnbrec_transit_switch *key;
 
-    struct hmap routes_ad = HMAP_INITIALIZER(&routes_ad);
+    struct hmap *routes_ad;
+    const struct icnbrec_transit_switch *t_sw;
     for (int i = 0; i < ic_lr->n_isb_pbs; i++) {
         isb_pb = ic_lr->isb_pbs[i];
         key = icnbrec_transit_switch_index_init_row(
             ctx->icnbrec_transit_switch_by_name);
         icnbrec_transit_switch_index_set_name(key, isb_pb->transit_switch);
-        ts_name = icnbrec_transit_switch_index_find(
-            ctx->icnbrec_transit_switch_by_name, key)->name;
+        t_sw = icnbrec_transit_switch_index_find(
+             ctx->icnbrec_transit_switch_by_name, key);
         icnbrec_transit_switch_index_destroy_row(key);
+        if (!t_sw) {
+            continue;
+        }
+        ts_name = t_sw->name;
+        routes_ad = shash_find_data(routes_ad_by_ts, ts_name);
+        if (!routes_ad) {
+            routes_ad = xzalloc(sizeof *routes_ad);
+            hmap_init(routes_ad);
+            shash_add(routes_ad_by_ts, ts_name, routes_ad);
+        }
 
         if (!extract_lsp_addresses(isb_pb->address, &ts_port_addrs)) {
             static struct vlog_rate_limit rl = VLOG_RATE_LIMIT_INIT(5, 1);
@@ -1626,12 +1667,10 @@ advertise_lr_routes(struct ic_context *ctx,
         }
         lrp_name = get_lrp_name_by_ts_port_name(ctx, isb_pb->logical_port);
         route_table = get_route_table_by_lrp_name(ctx, lrp_name);
-        build_ts_routes_to_adv(ctx, ic_lr, &routes_ad, &ts_port_addrs,
+        build_ts_routes_to_adv(ctx, ic_lr, routes_ad, &ts_port_addrs,
                                nb_global, route_table);
-        advertise_routes(ctx, az, ts_name, &routes_ad);
         destroy_lport_addresses(&ts_port_addrs);
     }
-    hmap_destroy(&routes_ad);
 }
 
 static void
@@ -1732,27 +1771,113 @@ route_run(struct ic_context *ctx,
     icsbrec_port_binding_index_destroy_row(isb_pb_key);
 
     struct ic_router_info *ic_lr;
+    struct shash routes_ad_by_ts = SHASH_INITIALIZER(&routes_ad_by_ts);
     HMAP_FOR_EACH_SAFE (ic_lr, node, &ic_lrs) {
-        advertise_lr_routes(ctx, az, ic_lr);
-        sync_learned_routes(ctx, az, ic_lr);
+        collect_lr_routes(ctx, ic_lr, &routes_ad_by_ts);
+        sync_learned_routes(ctx, ic_lr);
         free(ic_lr->isb_pbs);
         hmap_destroy(&ic_lr->routes_learned);
         hmap_remove(&ic_lrs, &ic_lr->node);
         free(ic_lr);
     }
+    struct shash_node *node;
+    SHASH_FOR_EACH (node, &routes_ad_by_ts) {
+        advertise_routes(ctx, az, node->name, node->data);
+        hmap_destroy(node->data);
+    }
+    shash_destroy_free_data(&routes_ad_by_ts);
     hmap_destroy(&ic_lrs);
 }
 
+/*
+ * This function implements a sequence number protocol that can be used by
+ * the INB end user to verify that ISB is synced with all the changes that
+ * are done be the user/AZs-controllers:
+ *
+ * Since we have multiple IC instances running in different regions
+ * we can't rely on one of them to update the ISB and sync that update
+ * to INB since other ICs can make changes in parallel.
+ * So to have a sequence number protocol working properly we must
+ * make sure that all the IC instances are synced with the ISB first
+ * and then update the INB.
+ *
+ * To guarantee that all instances are synced with ISB first, each IC
+ * will do the following steps:
+ *
+ * 1. when local ovn-ic sees that INB:nb_ic_cfg has updated we will set
+ *    the ic_sb_loop->next_cfg to match the INB:nb_ic_cfg and increment
+ *    the value of AZ:nb_ic_cfg and wait until we get confirmation from
+ *    the server.
+ *
+ * 2. once this IC instance changes for ISB are committed successfully
+ *    (next loop), the value of cur_cfg will be updated to match
+ *    the INB:nb_ic_cfg that indicate that our local instance is up to date
+ *    and no more changes need to be done for ISB.
+ *
+ * 3. validate that the AZ:nb_ic_cfg to match the INB:nb_ic_cfg.
+ *
+ * 4. Go through all the AZs and check if all have the same value of
+ *    AZ:nb_ic_cfg that means all the AZs are done with ISB changes and ISB are
+ *    up to date with INB, so we can set the values of ISB:nb_ic_cfg to
+ *    INB:nb_ic_cfg and INB:sb_ic_cfg to INB:nb_ic_cfg.
+ */
 static void
-ovn_db_run(struct ic_context *ctx)
+update_sequence_numbers(const struct icsbrec_availability_zone *az,
+                        struct ic_context *ctx,
+                        struct ovsdb_idl_loop *ic_sb_loop)
 {
-    const struct icsbrec_availability_zone *az = az_run(ctx);
-    VLOG_DBG("Availability zone: %s", az ? az->name : "not created yet.");
-
-    if (!az) {
+    if (!ctx->ovnisb_txn || !ctx->ovninb_txn) {
         return;
     }
 
+    const struct icnbrec_ic_nb_global *ic_nb = icnbrec_ic_nb_global_first(
+                                               ctx->ovninb_idl);
+    if (!ic_nb) {
+        ic_nb = icnbrec_ic_nb_global_insert(ctx->ovninb_txn);
+    }
+    const struct icsbrec_ic_sb_global *ic_sb = icsbrec_ic_sb_global_first(
+                                               ctx->ovnisb_idl);
+    if (!ic_sb) {
+        ic_sb = icsbrec_ic_sb_global_insert(ctx->ovnisb_txn);
+    }
+
+    if ((ic_nb->nb_ic_cfg != ic_sb->nb_ic_cfg) &&
+                          (ic_nb->nb_ic_cfg != az->nb_ic_cfg)) {
+        /* Deal with potential overflows. */
+        if (az->nb_ic_cfg == LLONG_MAX) {
+            icsbrec_availability_zone_set_nb_ic_cfg(az, 0);
+        }
+        ic_sb_loop->next_cfg = ic_nb->nb_ic_cfg;
+        ovsdb_idl_txn_increment(ctx->ovnisb_txn, &az->header_,
+                           &icsbrec_availability_zone_col_nb_ic_cfg, true);
+        return;
+    }
+
+    /* handle cases where accidentally AZ:ic_nb_cfg exceeds
+     * the INB:ic_nb_cfg.
+     */
+    if (az->nb_ic_cfg != ic_sb_loop->cur_cfg) {
+        icsbrec_availability_zone_set_nb_ic_cfg(az, ic_sb_loop->cur_cfg);
+        return;
+    }
+
+    const struct icsbrec_availability_zone *other_az;
+    ICSBREC_AVAILABILITY_ZONE_FOR_EACH (other_az, ctx->ovnisb_idl) {
+        if (other_az->nb_ic_cfg != az->nb_ic_cfg) {
+            return;
+        }
+    }
+    /* All the AZs are updated successfully, update SB/NB counter. */
+    if (ic_nb->nb_ic_cfg != ic_sb->nb_ic_cfg) {
+        icsbrec_ic_sb_global_set_nb_ic_cfg(ic_sb, az->nb_ic_cfg);
+        icnbrec_ic_nb_global_set_sb_ic_cfg(ic_nb, az->nb_ic_cfg);
+    }
+}
+
+static void
+ovn_db_run(struct ic_context *ctx,
+           const struct icsbrec_availability_zone *az)
+{
     ts_run(ctx);
     gateway_run(ctx, az);
     port_binding_run(ctx, az);
@@ -1891,6 +2016,31 @@ update_ssl_config(void)
     }
 }
 
+static void
+update_idl_probe_interval(struct ovsdb_idl *ovn_sb_idl,
+                          struct ovsdb_idl *ovn_nb_idl,
+                          struct ovsdb_idl *ovn_icsb_idl,
+                          struct ovsdb_idl *ovn_icnb_idl)
+{
+    const struct nbrec_nb_global *nb = nbrec_nb_global_first(ovn_nb_idl);
+    int interval = -1;
+    if (nb) {
+        interval = smap_get_int(&nb->options, "ic_probe_interval", interval);
+    }
+    set_idl_probe_interval(ovn_sb_idl, ovnsb_db, interval);
+    set_idl_probe_interval(ovn_nb_idl, ovnnb_db, interval);
+
+    const struct icnbrec_ic_nb_global *icnb =
+        icnbrec_ic_nb_global_first(ovn_icnb_idl);
+    int ic_interval = -1;
+    if (icnb) {
+        ic_interval = smap_get_int(&icnb->options, "ic_probe_interval",
+                                   ic_interval);
+    }
+    set_idl_probe_interval(ovn_icsb_idl, ovn_ic_sb_db, ic_interval);
+    set_idl_probe_interval(ovn_icnb_idl, ovn_ic_nb_db, ic_interval);
+}
+
 int
 main(int argc, char *argv[])
 {
@@ -1906,7 +2056,7 @@ main(int argc, char *argv[])
     service_start(&argc, &argv);
     parse_options(argc, argv);
 
-    daemonize_start(false);
+    daemonize_start(false, false);
 
     char *abs_unixctl_path = get_abs_unix_ctl_path(unixctl_path);
     retval = unixctl_server_create(abs_unixctl_path, &unixctl);
@@ -2085,12 +2235,23 @@ main(int argc, char *argv[])
                                   &icsbrec_route_col_transit_switch,
                                   &icsbrec_route_col_availability_zone);
 
+    unixctl_command_register("nb-connection-status", "", 0, 0,
+                             ovn_conn_show, ovnnb_idl_loop.idl);
+    unixctl_command_register("sb-connection-status", "", 0, 0,
+                             ovn_conn_show, ovnsb_idl_loop.idl);
+    unixctl_command_register("ic-nb-connection-status", "", 0, 0,
+                             ovn_conn_show, ovninb_idl_loop.idl);
+    unixctl_command_register("ic-sb-connection-status", "", 0, 0,
+                             ovn_conn_show, ovnisb_idl_loop.idl);
+
     /* Main loop. */
     exiting = false;
     state.had_lock = false;
     state.paused = false;
     while (!exiting) {
         update_ssl_config();
+        update_idl_probe_interval(ovnsb_idl_loop.idl, ovnnb_idl_loop.idl,
+                                  ovnisb_idl_loop.idl, ovninb_idl_loop.idl);
         memory_run();
         if (memory_should_report()) {
             struct simap usage = SIMAP_INITIALIZER(&usage);
@@ -2151,13 +2312,28 @@ main(int argc, char *argv[])
                 ovsdb_idl_has_ever_connected(ctx.ovnsb_idl) &&
                 ovsdb_idl_has_ever_connected(ctx.ovninb_idl) &&
                 ovsdb_idl_has_ever_connected(ctx.ovnisb_idl)) {
-                ovn_db_run(&ctx);
+                const struct icsbrec_availability_zone *az = az_run(&ctx);
+                VLOG_DBG("Availability zone: %s", az ? az->name :
+                                               "not created yet.");
+                if (az) {
+                    ovn_db_run(&ctx, az);
+                    update_sequence_numbers(az, &ctx, &ovnisb_idl_loop);
+                }
             }
 
-            ovsdb_idl_loop_commit_and_wait(&ovnnb_idl_loop);
-            ovsdb_idl_loop_commit_and_wait(&ovnsb_idl_loop);
-            ovsdb_idl_loop_commit_and_wait(&ovninb_idl_loop);
-            ovsdb_idl_loop_commit_and_wait(&ovnisb_idl_loop);
+            int rc1 = ovsdb_idl_loop_commit_and_wait(&ovnnb_idl_loop);
+            int rc2 = ovsdb_idl_loop_commit_and_wait(&ovnsb_idl_loop);
+            int rc3 = ovsdb_idl_loop_commit_and_wait(&ovninb_idl_loop);
+            int rc4 = ovsdb_idl_loop_commit_and_wait(&ovnisb_idl_loop);
+            if (!rc1 || !rc2 || !rc3 || !rc4) {
+                VLOG_DBG(" a transaction failed in: %s %s %s %s",
+                         !rc1 ? "nb" : "", !rc2 ? "sb" : "",
+                         !rc3 ? "ic_nb" : "", rc4 ? "ic_sb" : "");
+                /* A transaction failed. Wake up immediately to give
+                 * opportunity to send the proper transaction
+                 */
+                poll_immediate_wake();
+            }
         } else {
             /* ovn-ic is paused
              *    - we still want to handle any db updates and update the
@@ -2233,7 +2409,7 @@ ovn_ic_resume(struct unixctl_conn *conn, int argc OVS_UNUSED,
 {
     struct ic_state *state = state_;
     state->paused = false;
-
+    poll_immediate_wake();
     unixctl_command_reply(conn, NULL);
 }
 
